#include <arpa/inet.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include "interface.h"

void spawn_server_instance() {
  pid_t pid = fork();
  if (pid == -1) {
    err(EXIT_FAILURE, "Failed to fork for the server process");
  } else if (pid == 0) {
    char *argv[] = {"/home/server/spawn_server", 0};
    if (execv(argv[0], &argv[0])) {
      err(EXIT_FAILURE, "Failed to spawn spawn_server process");
    }
  }
  usleep(100000);
}

static char *shared_mem = NULL;
void open_shared_memory() {
  int fd;
  do {
    fd = shm_open(SHM_NAME, O_RDONLY, 0);
    if (fd == -1) {
      if (errno != ENOENT) {
        err(EXIT_FAILURE, "Failed to open shared memory file %s", SHM_NAME);
      } else {
        sched_yield();
      }
    }
  } while (fd == -1);

  shared_mem = mmap(SHM_ADDR, SHM_SIZE, PROT_READ,
                    MAP_SHARED | MAP_POPULATE | MAP_FIXED, fd, 0);
  if (shared_mem == MAP_FAILED) {
    err(EXIT_FAILURE, "Failed to mmap shared memory");
  }
}

void send_request(int sock_fd, const struct sockaddr_in *dest_addr,
                  unsigned char day_of_month) {
  ssize_t bytes =
      sendto(sock_fd, &day_of_month, 1, 0, (const struct sockaddr *)dest_addr,
             sizeof(*dest_addr));
  if (bytes == -1) {
    err(EXIT_FAILURE, "Failed to send QOTD request");
  }
}

int main() {
  spawn_server_instance();
  open_shared_memory();

  int sock = socket(AF_INET, SOCK_DGRAM, 0);
  if (sock == -1) {
    err(EXIT_FAILURE, "Failed to open socket");
  }

  const struct sockaddr_in serveraddr = {.sin_family = AF_INET,
                                         .sin_port = htons(17),
                                         .sin_addr.s_addr =
                                             htonl(INADDR_LOOPBACK)};
  time_t t;
  time(&t);
  struct tm *tm = localtime(&t);
  send_request(sock, &serveraddr, tm->tm_mday - 1);

  usleep(100000);
  printf("%s\n", shared_mem);
}
