#include <arpa/inet.h>
#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <unistd.h>

#include "interface.h"

struct data {
  char day_of_month_to_quote_idx[32];
  char *shared_mem;
  char flag_path[24];
  char flag[64];
  unsigned char days_in_month;
} __attribute__((aligned(64)));

struct data data = {
    .day_of_month_to_quote_idx = {30, 29, 9,  22, 25, 7,  18, 28, 12, 23, 3,
                                  14, 2,  10, 16, 0,  1,  4,  6,  19, 15, 11,
                                  5,  17, 24, 27, 21, 26, 13, 20, 8},
    .shared_mem = NULL,
    .flag_path = "/home/server/flag.txt",
    .flag = {0},
    .days_in_month = 31};

void read_flag() {
  FILE *file = fopen(data.flag_path, "r");
  if (!file) {
    err(EXIT_FAILURE, "Failed to read %s", data.flag_path);
  }

  fread(data.flag, 1, sizeof(data.flag), file);

  fclose(file);
}

static char *shared_mem = NULL;
void setup_shared_memory() {
  int fd;
  do {
    fd = shm_open(SHM_NAME, O_RDWR | O_CREAT | O_EXCL | O_TRUNC,
                  S_IRUSR | S_IWUSR | S_IROTH);

    if (fd == -1) {
      if (errno == EEXIST) {
        if (shm_unlink(SHM_NAME) == -1) {
          err(EXIT_FAILURE, "Failed to unlink shared memory %s", SHM_NAME);
        }
      }
    }
  } while (fd == -1);

  if (ftruncate(fd, SHM_SIZE) == -1) {
    err(EXIT_FAILURE, "Failed to resize shared memory file");
  }

  shared_mem = mmap(SHM_ADDR, SHM_SIZE, PROT_READ | PROT_WRITE,
                    MAP_SHARED | MAP_POPULATE | MAP_FIXED, fd, 0);
  if (shared_mem == MAP_FAILED) {
    err(EXIT_FAILURE, "Failed to mmap shared memory");
  }
}

static size_t quote_cursor = PAGE_SIZE;
void load_quotes() {
  FILE *fp = fopen("/home/server/quotes.txt", "r");
  if (!fp) {
    err(EXIT_FAILURE, "Failed to open quotes.txt");
  }

  size_t line_length = 0;
  char *line = NULL;
  ssize_t read;

  size_t nr_of_quotes = 0;
  while (read = getline(&line, &line_length, fp), read != -1) {
    memcpy(&shared_mem[quote_cursor], line, line_length);
    quote_cursor += PAGE_SIZE;
    ++nr_of_quotes;
  }
  assert(nr_of_quotes <= sizeof(data.day_of_month_to_quote_idx));

  free(line);
  fclose(fp);
}

void copy_quote(unsigned char day_of_month) {
  __builtin_ia32_clflush((const void *)&data.days_in_month);

  if (day_of_month < data.days_in_month) {
    memcpy(
        &shared_mem[0],
        &shared_mem[data.day_of_month_to_quote_idx[day_of_month] * PAGE_SIZE],
        128);
  }
}

int main() {
  if (prctl(PR_SET_PDEATHSIG, SIGTERM) == -1) {
    err(EXIT_FAILURE, "Failed to set parent death signal");
  }
  if (getppid() == 1) {
    err(EXIT_FAILURE, "Parent died before server could spawn");
  }

  read_flag();
  setup_shared_memory();
  load_quotes();

  int fd = socket(PF_INET, SOCK_DGRAM, 0);
  if (fd == -1) {
    err(EXIT_FAILURE, "Failed to create socket");
  }

  const struct sockaddr_in serveraddr = {.sin_family = AF_INET,
                                         .sin_port = htons(17),
                                         .sin_addr.s_addr =
                                             htonl(INADDR_LOOPBACK)};
  if (bind(fd, (const struct sockaddr *)&serveraddr, sizeof(serveraddr))) {
    err(EXIT_FAILURE, "Failed to bind server to port");
  }

  while (1) {
    unsigned char day_of_month;
    ssize_t bytes = recv(fd, &day_of_month, 1, 0);
    if (bytes == -1) {
      err(EXIT_FAILURE, "Failed to recv");
    }
    copy_quote(day_of_month);
    sched_yield();
  }

  close(fd);
}
